package se.chalmers.cs.gf.translate;

import se.chalmers.cs.gf.GFException;
import se.chalmers.cs.gf.gfcutil.*;
import se.chalmers.cs.gf.parse.*;
import se.chalmers.cs.gf.parse.chart.*;
import se.chalmers.cs.gf.linearize.*;
import se.chalmers.cs.gf.util.FileUtil;

import java.io.*;
import java.util.*;


/**
 *  Loads parsers, linearizers etc. and creates Translators from them.
 */
public class TranslatorFactory {

        private static List<Object> instantiateAll(String[] classes) {
                List<Object> cs = new LinkedList<Object>();
                for (String n : classes) {
                        try {
                                Class<?> c = Class.forName(n);
                                cs.add(c.newInstance());
                        } catch (ClassNotFoundException ex) {
                                System.err.println(ex.toString());
                        } catch (InstantiationException ex) {
                                System.err.println(ex.toString());
                        } catch (IllegalAccessException ex) {
                                System.err.println(ex.toString());
                        }
                }
                return cs;
        }

        private static Reader readResource(String name) throws IOException {
                return FileUtil.readResource(Translator.class, name, "UTF-8");
        }

        /**
         *  Create a translator by reading properties from the given resource.
         */
        public static Translator createTranslator(String propFile) throws IOException {
                Properties prop = FileUtil.readPropertiesResource(Translator.class, propFile);
                return createTranslator(prop);
        }

        /**
         *  Create a translator from a set of properties.
         */
        public static Translator createTranslator(Properties prop) throws IOException {
                String name = FileUtil.getProp(prop, "name");
                String cfgm = FileUtil.getProp(prop, "cfgm");
                String[] xps = FileUtil.getListProp(prop, "extraparsers");
                String gfcm = FileUtil.getProp(prop, "gfcm");
                String[] xls = FileUtil.getListProp(prop, "extralinearizers");

                return createTranslator(name, cfgm, gfcm, xps, xls);
        }

        /**
         *  Create a translator by reading grammars from the given resources.
         *  @param name The name of the translator.
         *  @param cfgm Resource to read the CFGM (parsing) grammar from.
         *  @param gfcm Resource to read the GFCM (linearization) grammar from.
         */
        public static Translator createTranslator(String name, String cfgm, String gfcm) 
                throws IOException {

                return createTranslator(name, cfgm, gfcm, new String[0], new String[0]);
        }

        /**
         *  Create a translator by reading grammars from the given resources.
         *  @param name The name of the translator.
         *  @param cfgm Resource to read the CFGM (parsing) grammar from.
         *  @param gfcm Resource to read the GFCM (linearization) grammar from.
         *  @param extraparsers Class names of extra parsers to load
         *  @param extralinearizers Class names of extra linearizers to load
         */
        public static Translator createTranslator(String name, 
                                                  String cfgm, 
                                                  String gfcm, 
                                                  String[] extraparsers, 
                                                  String[] extralinearizers) 
                throws IOException {

//                System.err.println("Creating parsers from " + cfgm + "...");
                Parsers ps = CFGParsers.readParsers(readResource("/"+cfgm));

                ps.addParser(new ParseAll(ps));

                for (Object o : instantiateAll(extraparsers)) {
                        Parser p = (Parser)o;
//                        System.err.println("Loaded extra parser " + p.getName());
                        ps.addParser(p);
                }

//                System.err.println("Creating linearizers from " + gfcm + "...");
                GFCMModules ms = GFCMModules.loadGFCM(readResource("/"+gfcm));
                Linearizers ls = new GFCMLinearizers(ms);

                ls.addLinearizer(new LinearizeAll(ls));

                for (Object o : instantiateAll(extralinearizers)) {
                        Linearizer l = (Linearizer)o;
//                        System.err.println("Loaded extra linearizer " + l.getName());
                        ls.addLinearizer(l);
                }

                GFCModule absModule = ms.getMainAbstractModule();
                TypeAnnotator annotator = new TypeAnnotator(absModule);

                return new Translator(name, ps, annotator, ls);
        }

}
